# /r2/martins_sketch_talk.rb

require 'sketchup'
dir = File.expand_path( File.dirname(__FILE__) )
load File.join( dir, 'movable_ci.rb' )
load File.join( dir, 'sketch_talk_classes.rb' )
load File.join( dir, 'transformable_ci.rb' )

$sketch_talk_conductor = nil
$sketch_talk_fps = 24.0 # default frames per second, must be a float!
$radians_per_degree = Math::PI / 180.0

# SketchTalk Functions ------------------------------------------

def all()

	model = Sketchup.active_model()
    model.selection().clear()
    ents = []
	for e in model.entities()
		ents.push( e ) if e.layer().visible?()
	end
	model.selection().add( ents )
    
    $sketch_talk_logFile.puts( 'all' ) if $sketch_talk_logFile
    
    return model.selection()
    
end # of all()

def box( near, far, pushpull )
    b = Box.new( near, far, pushpull )
    b.draw()
    return b
end # of box()

def c( nsides, center, radius, plane_or_normal )
    circ = Circle.new( nsides, center, radius, plane_or_normal )
    circ.draw( Sketchup.active_model().entities() )
    return circ
end # of c()

def col( nsides, bottom_center, radius, 
        plane_or_normal, pushpull_distance  ) 
    # this is a circle plus a pushpull height
    
    column = Column.new( nsides, bottom_center, 
        radius, plane_or_normal, pushpull_distance  ) 
    column.draw( Sketchup.active_model().entities() )
    return column

end # of col() 

def cz() # Ctrl+Z from the Ruby Console
    Sketchup.send_action( 'editUndo:' )
end

def del()
	model = Sketchup.active_model()
	sel = model.selection()
	ents = model.entities()
    
	sel_ents = []
	for e in sel do sel_ents.push( e ) end
	
    ents.erase_entities( sel_ents )

    $sketch_talk_logFile.puts( 'del' ) if $sketch_talk_logFile
    
end # of del()

def donut( near, far, pushpull, thickness )
    d = Doughnut.new( near, far, pushpull, thickness )
    d.draw()
    return d
end # of donut()

def draw()
    Sketchup.active_model().active_view().invalidate()
end

def face( *args )
    return add_face_group( args )
end # of face

def fs( *args ) # File/Save
end

def g( name )
	model = Sketchup.active_model 
	return make_component( model.selection(), name )
end # of g()

def go()
    if $sketch_talk_conductor == nil
        UI.messagebox 'Command "go": no animations specified.'
        return
    end

    $sketch_talk_conductor.start()
    return true
end

def i( pathname ) # file/import

    model = Sketchup.active_model

    pathname += '.skp' unless /\.skp$/.match( pathname )
    defs = model.definitions
    defi = defs.load( pathname )

    inst = model.entities.add_instance( defi, Geom::Transformation.new() )
    model.selection.add( inst )
    return inst

end # of i()

def l( *args )
	ents = Sketchup.active_model.entities 
outs 'in l, args = ' + args.inspect()
	return ents.add_edges( args )
end # of l()

def lst()
    load File.dirname( __FILE__ ) + '/martins_sketch_talk2.rb'
end # of lst()

def m( *args )
    unless ( args.length > 0 ) && ( (args[0].is_a?( Sketchup::ComponentInstance )) || (args[0].is_a?( Sketchup::Group )) )
        UI.messagebox( "First argument of \"m()\" must be a Sketchup::ComponentInstance or Sketchup::Group.\n" +
            'Use "xm()" to translate by r, g, b values.') 
        return
    end
    
    case args.length()
        when 2 then move( args ) # static move
        when 3 then movie( args ) # animated move
        when 4 then movie2( args ) # movie entered on movie stack
        when 1 then xmove( args[0] )
        else UI.messagebox( 'm args must be: ( transformation ) or ( comp, vec ) or ( comp, vec, ntimes ) or ( comp, vec, start, stop ).' )
    end
end # of m()

def mc( comp, dist, times )
	move_copy( comp, dist, times )
end # of mc()

def mci( inst )
    return MovableCI.new( inst )
end # of mci()

def n()
    all()
    del()
    $sketch_talk_conductor = nil
end # of n()

def new()
    Sketchup.file_new()
    Sketchup.send_action( 'selectSelectionTool:' )
end # of new()

def none()
    Sketchup.active_model().selection().clear()
end # of none()

def o()
    sel = Sketchup.active_model().selection()
    return sel[0].transformation().origin() if sel.length() && 
        sel[0].is_a?( Sketchup::ComponentInstance )
    return Geom::Point3d.new( 0, 0, 0 )
end # of o()
    
def ogb( size ) # rectangle in gb plane from origin
    r = Rectangle.new( [0,0,0], [0,size,size] )
    r.draw()
    return r
end # of ogb()

def orb( size )
    r = Rectangle.new( [0,0,0], [size, 0, size] )
    r.draw()
    return r
end # of orb()

def org( size )
    r = Rectangle.new( [0,0,0], [size,size,0] )
    r.draw()
    return r
end # of org()

def p ( near, far, height )

    near = near.to_a() # may be Geom::Point3d
    far = far.to_a()
    
    corners = [
        near,
        [ far[0], near[0], near[2] ],
        [ far[0], far[1], near[2] ],
        [ near[0], far[1], near[2] ]
    ]
    center = [
        ( near[0] + far[0] ) / 2.0,
        ( near[1] + far[1] ) / 2.0,
        near[ 2 ]
    ]
    peak = [ center[0], center[1], center[2] + height ]
    
    ents = Sketchup.active_model().entities()
    grp = ents.add_group()
    
    ents = grp.entities()
    
    ents.add_face( corners )
    ents.add_face( corners[0], corners[1], peak )
    ents.add_face( corners[1], corners[2], peak )
    ents.add_face( corners[2], corners[3], peak )
    ents.add_face( corners[3], corners[0], peak )
    
    sel = Sketchup.active_model().selection()
    sel.clear()
    sel.add( grp )
    return grp
    
end # of p()

def q( *args )
=begin
Qrotate component around axis set by pt plus plane_or_normal, angle degrees.

Arguments:
"comp" is any ComponentInstance. 
"pt" is an [r,g,b] array. 
"plane_or_normal" is "rg", "gb", "rb" or another [r,g,b] array. 
"ntimes" is an optional repeat specification
=end

    qrotate( *args )

end # of q()

def r( *args )
    if args.length == 2
        r = Rectangle.new( args[0], args[1] )
        r.draw()
        return r
    elsif args.length == 4
        return face( args[0], args[1], args[2], args[3] )
    else
        raise "Command r requires two or four points."
    end
end # of r()

def s( *args )
    case args.length
        when 1 then
            t = Geom::Transformation.scaling( args[0] )
        when 2 then
            t = Geom::Transformation.scaling( args[0], args[1] )
        when 3 then
            t = Geom::Transformation.scaling( args[0], args[1], args[2] )
        when 4 then
            t = Geom::Transformation.scaling( args[0], args[1], args[2], args[3] )
        else
            UI.messagebox( "Scale arg(s) must be one of:\n",
            "\tglobal_scale (about origin)\n",
            "\tpoint, glogal_scale\n",
            "\trscale, gscale, bscale (about origin)\n",
            "\tpoint, rscale, gscale, bscale." )
            return
    end # of cases

    sel().transform!( t )
    draw()
end

def sel( )
    # returns reference to first selected item
    return Sketchup.active_model().selection()[0]
end

def tci( inst )
    return TransformableCI.new( inst )
end # of tci()

def tri( p1, p2, p3 )
    return add_face_group( p1, p2, p3 )
end # of tri()

def wd( msg )
    web_dialog( msg )
end

def where()
    t = sel().transformation().to_a()
    return [ t[12], t[13], t[14] ]
end # of where()
    
def xf( *args ) # display selected object's transformation matrix
=begin
No args? Launch dialog showing selected object's transformation matrix

One arg? Arg is an xform matrix. Apply it to selection, redraw and display it in dialog.
=end
    if args.length == 0 
        xform( "
        <html>
        <body>
        <table align=left bgcolor=#f0f0ff border=1 cellpadding=3>
            <tr>
                <td id=c0 align=right> </td>
                <td id=c1 align=right> </td>
                <td id=c2 align=right> </td>
                <td id=c3 align=right> </td>
            </tr>
            <tr>
                <td id=c4 align=right> </td>
                <td id=c5 align=right> </td>
                <td id=c6 align=right> </td>
                <td id=c7 align=right> </td>
            </tr>
            <tr>
                <td id=c8 align=right> </td>
                <td id=c9 align=right> </td>
                <td id=c10 align=right> </td>
                <td id=c11 align=right> </td>
            </tr>
            <tr>
                <td id=c12 align=right> </td>
                <td id=c13 align=right> </td>
                <td id=c14 align=right> </td>
                <td id=c15 align=right> </td>
            </tr>

        <script type='text/javascript'>
            function set_vals( vals ) {
            for ( var i = 0; i < 16; i++ ) {

                td = document.getElementById( 'c' + i );
                td.innerHTML = vals[i];
            }
        }
        vals =
        " )
    else # end of display new dialog showing xform matrix
        sel.move!( args[0] )
        draw()
        xf()
    end

end # of xf()

def xm( r, g, b ) # move the currently selected component by specified amounts
    trans = translate( sel.transformation, r, g, b )
    sel.move!( trans )
    draw()
end # of xm()


def xpld( arg )
    if ( arg.is_a? Sketchup::Group ) || 
        ( arg.is_a? Sketchup::ComponentInstance )
        arg.explode()
    else
        arg.group.explode()
    end
end
    
def ze()
    Sketchup.send_action( 'viewZoomExtents:' )
end # of ze()

# Support Functions ---------------------------------------------

def add_face_group( args ) # [entities,] points
=begin
Adds a group to the Entities specified as args[0] or, if not specified, to Sketchup.active_model().entities(). Adds face specified by points to that group. Selects the group and returns it.
=end

    model = Sketchup.active_model()

    if args[0].is_a?( Sketchup::Entities )
        ents = args.shift()
    else
        ents = model.entities()
    end
    
    grp = ents.add_group()
    face = grp.entities().add_face( args )
    orient( face )
    model.selection().add( grp )
    
    return grp
    
end # of add_face_group()

def find_face( ents )

    for e in ents do return e if e.is_a?( Sketchup::Face ) end
    
end # of find_face()
    
def make_component( entities, name )

=begin
Given an array of entities and a name, returns a named component containing the entities.
=end

    model = Sketchup.active_model() 
	ents = model.entities()

	g = ents.add_group( entities ) # group all entities
	inst = g.to_component() # convert group to component instance
	inst.definition().name = name
    
    model.selection().add( inst )

	return inst

end # of make_component()

def make_normal( plane_or_normal )

    if plane_or_normal.is_a?( String )
        case plane_or_normal
            when 'rg' then normal = [0,0,1]
            when 'rb' then normal = [0,1,0]
            when 'gb' then normal = [1,0,0]
            else
                raise "Plane must be 'rg', 'rb', 'gb' or a normal vector."
        end
    else
        normal = plane_or_normal
    end
    
    return normal

end # of make_normal()

def move( args ) # static move

    inst = args[0]
    if inst.is_a?( Rectangle )
        inst.move( args[1] )
    else
        trans = translate( inst.transformation, args[1] )
        inst.move!( trans )
        draw()
    end
    
end # of move()

def move_copy( component, distance, number_of_copies )

=begin
Creates an outer array, moving a component "distance" as many times as specified.

"distance" is a vector, an array of [r, g, b].

"number_of_copies" is like the "Nx" in the VCB after a Move/Copy.
For 15 steps, you make 14 copies.
=end

    ents = Sketchup.active_model.entities

    defi = component.definition # the original component's definition
    trans = component.transformation # the original's transformation
    pt = trans.origin # the original's location, a Point3d

    for i in ( 1..number_of_copies )
        pt += distance
            # add vector to Point3d getting new Point3d
        trans = Geom::Transformation.new( pt )
            # create new Transformation at the new Point3d
        ents.add_instance( defi, trans )
            # add another instance at the new Point3d
    end

end # of move_copy()

def movie( args ) # animated move

    inst = args[0]
    vec = args[1]
    nsecs = args[2]
    
    nframes = nsecs * $sketch_talk_fps # $sketch_talk_fps is a float
    frame_vec = [ vec[0]/nframes, vec[1]/nframes, vec[2]/nframes ]
    mover = Mover.new( frame_vec, nframes, inst )
    anim = Sketchup.active_model.active_view.animation = mover
    
end # of movie()

def movie2( args ) # movie, the right way
    # args are reference, vector, start, stop
    # object pointed to by reference will move vector
    # movement begins with start and continues up to, but not including, stop

    $sketch_talk_conductor = Conductor.new() if $sketch_talk_conductor == nil
    
    $sketch_talk_conductor.register( args ) 

end # of movie2()

class Mover
    attr_reader :done

   def initialize( vector, nframes, inst )
        @vector = vector
        @nframes = nframes
        @inst = inst
        @count = 0
        @start = Time.now
        @done = false
        @delay = 1.0 / $sketch_talk_fps
    end
    
    def nextFrame( view )
        trans = translate( @inst.transformation, @vector )
        @inst.move!( trans )
        view.show_frame( @delay )
        @count += 1
        puts Time.now() - @start if @count == @nframes
        return @count < @nframes
    end
    
    def stop
        @done = true
    end
    
    def inspect()
        return 'Mover:' +
            ' @vector=' + format( '%.2f, %.2f, %.2f', 
                @vector[0], @vector[1], @vector[2] ) +
            ' @nframes=' + @nframes.to_s() +
            ' @delay=' + format( '%.2f', @delay )
    end

end # of class Mover

def orient( face )
=begin
Orthogonal faces face the positive end of the perpendicular axis.
Non-orthogonal faces with a non-zero blue normal face the positive end of the blue axis.
Non-orthogonal faces parallel to the blue axis face the positive end of the green axis.
=end

    norm = face.normal()

    # orthogonal face pointing toward positive end of perpendicular?
    return if norm == ( [0,0,1] ) || ( norm == [0,1,0] ) || ( norm == [1,0,0] )
    
    # orthogonal face pointing toward negative end of perpendicular?
    if norm == ( [0,0,-1] ) || ( norm == [0,-1,0] ) || ( norm == [-1,0,0] )
        face.reverse!()
        return
    end
    
    # non-orthogonal, not parallel to blue axis?
    return if norm.z > 0
    if norm.z < 0
        face.reverse!()
        return
    end
    
    # non-orthogonal, parallel to blue axis
    face.reverse!() if norm.y < 0
    return
    
end # of orient()

def qrotate( *args ) # ( comp, pt, plane_or_normal, angle[, ntimes] )
=begin
Qrotate component around axis set by pt plus plane_or_normal, by angle degrees.

"comp" is any ComponentInstance. 
"pt" is an [r,g,b] array. 
"plane_or_normal" is "rg", "gb", "rb" or another [r,g,b] array. 
=end

    if args.length == 5
        ntimes = args.pop()
        id = UI.start_timer( 1.0/$sketch_talk_fps, true ) {
            qrotate( args[0], args[1], args[2], args[3] )
            ntimes -= 1
            UI.stop_timer( id ) if ntimes == 0
        }
        return
    end
    
    comp = args[0]

    unless comp.is_a?( Sketchup::ComponentInstance )
        UI.messagebox( 
            'Qrotate: first argument must be a Sketchup::ComponentInstance' )
        return
    end

    normal = make_normal( args[2] )
    angle = args[3] * $radians_per_degree
    
    trans = Geom::Transformation.rotation( args[1], normal, angle )
    comp.transform!( trans )
    draw()

end # of qrotate()

def translate( *args ) # add a translation vector to a transformation
=begin
May be called with a transformation and a vector, or with a transformation and r, g, b values.
=end
    trans = args[0]
    if args.length == 2
        vec = args[1]
        r = vec[0]; g = vec[1]; b = vec[2]
    else
        r = args[1]; g = args[2]; b = args[3]
    end
    arr = trans.to_a()
    arr[12] += r; arr[13] += g; arr[14] += b
    return Geom::Transformation.new( arr )
    
end # of translate()

def web_dialog( msg )
    UI.messagebox( msg )
end
    
def xform( html )

    model = Sketchup.active_model()
    sel = model.selection()
    
    if sel.length() == 0
        puts 'Nothing selected.'
        return
    end

    thing = sel[0]

    if thing.is_a?( Sketchup::ComponentInstance ) ||
            thing.is_a?( Sketchup::Group )
        trans = thing.transformation()
    else
        puts 'model.selection()[0] is not a Group or ComponentInstance.'
        return
    end
    
    # we don't really want stuff like 0.707106781186548
    nums = trans.to_a()
    snums = []
    for i in 0..2 do snums.push( sprintf('%6.3f', nums[i]) ) end
    snums.push( nums[3].to_s() )
    for i in 4..6 do snums.push( sprintf('%6.3f', nums[i]) ) end
    snums.push( nums[7].to_s() )
    for i in 8..10 do snums.push( sprintf('%6.3f', nums[i]) ) end
    snums.push( nums[11].to_s() )
    for i in 12..14 do snums.push( sprintf('%6.3f', nums[i]) ) end
    snums.push( nums[15].to_s() )
    
    html += snums.inspect()
    html += "
        set_vals( vals );
    </script>

</body>
</html>"
    wd = UI::WebDialog.new( "Transformation Matrix", true, 
			'xform', 350, 200, 0, 0, true )

    wd.set_html( html )

    wd.show()

end # of xform()

def xmove( trans )
    sel.transform!( trans )
    draw()
end

# Test Code -----------------------------------------------------

# timer = Timer.new( 1.0, Timer_test.new() )
# timer.start()

# e = Sketchup.active_model.entities[0]
# puts 'Mover has e?: ' + Mover.has( e ).to_s # false
# i = Mover.new( e )
# puts 'Mover has e?: ' + Mover.has( e ).to_s # true
# puts 'Mover.value(e): ' + Mover.value( e ).to_s
# puts 'Mover.value(e).inst.definition.name: ' +            
# Mover.value(e).inst.definition.name
# Mover.delete( e )
# puts 'Mover has e?: ' + Mover.has( e ).to_s # false

# m e, [10,0,5], 2

# end of /r2/martins_sketch_talk2.rb